import builtins
from inspect import currentframe

bprint = builtin_print = builtins.print


def debug(*args, condition=True):
    frame = currentframe().f_back
    filepath = _normpath(frame.f_globals["__file__"])
    lineno = frame.f_lineno
    source = '{}:{}'.format(filepath, lineno)
    if condition:
        builtin_print(source, '[LKDEBUG]', *args)


def _normpath(path: str) -> str:
    return path.replace('\\', '/').rstrip('/')



from __future__ import annotations

import typing as t


class T:  # Typehint
    from .markup import T as _TMarkup  # noqa
    
    FrameId = str
    MarkupPos = int
    Markup = str
    
    Cache = t.Dict[FrameId, t.TypedDict('_SubDict0', {
        'markup_pos'   : MarkupPos,  # noqa
        'marks_meaning': _TMarkup.MarksMeaning,  # noqa
        'info'         : t.Dict[Markup, dict]
    })]


class LoggingCache:
    _cache: T.Cache
    
    def __init__(self):
        from collections import defaultdict
        self._cache = defaultdict(lambda: {
            'markup_pos'   : None,
            'marks_meaning': {},
            'info'         : {},
        })
    
    def clear_cache(self):
        self._cache.clear()
    
    # -------------------------------------------------------------------------
    
    def get_markup_pos(self, frame_id: T.FrameId) -> T.MarkupPos | None:
        return self._cache[frame_id].get('markup_pos', None)
    
    def is_cached(self, frame_id: T.FrameId, markup: T.Markup) -> bool:
        return (
                frame_id in self._cache and
                markup in self._cache[frame_id]['info']
        )
    
    def get_cache(self, frame_id: T.FrameId, markup: T.Markup) -> dict:
        # suggest checking `self.is_cached` before calling this method.
        return self._cache[frame_id]['info'][markup]
    
    # -------------------------------------------------------------------------
    
    def record_markup_pos(self, frame_id: T.FrameId, pos: T.MarkupPos) -> None:
        self._cache[frame_id]['markup_pos'] = pos
    
    def store_info(self, frame_id: T.FrameId, markup: T.Markup, info) -> None:
        self._cache[frame_id]['info'][markup] = info



from __future__ import annotations

import os
import sys


def normpath(path: str) -> str:
    return path.replace('\\', '/').rstrip('/')


class PathHelper:
    __project_root: str
    __external_libs: dict[str, str] | None
    
    def __init__(self):
        self._cwd = os.getcwd()
        self.__project_root = self._find_project_root()
        self.__external_libs = None
    
    @staticmethod
    def _find_project_root() -> str:
        """ find personal-like project root.

        proposals:
            1. backtrack from current working dir to find a folder which has
               one of '.idea', '.git', etc. folders.
            2. iterate paths in sys.path, to find a folder which is parent of
               current working dir. (adopted)
               if there are more than one, choose which is shortest.
               if there is none, return current working dir.
        """
        cwd = normpath(os.getcwd())
        paths = tuple(
            x for x in map(normpath, map(os.path.abspath, sys.path))
            if cwd.startswith(x) and os.path.isdir(x)
        )
        if len(paths) == 0:
            return cwd
        elif len(paths) == 1:
            return paths[0]
        else:
            return min(paths, key=lambda x: len(x))
    
    # -------------------------------------------------------------------------
    
    @property
    def external_libs(self) -> dict[str, str]:
        if self.__external_libs is None:
            self.__external_libs = self._indexing_external_libs()
        return self.__external_libs
    
    def is_external_lib(self, filepath: str) -> bool:
        return not filepath.startswith(self.__project_root)
    
    @staticmethod
    def _indexing_external_libs() -> dict[str, str]:
        """
        return:
            dict[str path, str lib_name]
        """
        out = {}
        
        for path in reversed(sys.path):
            if not os.path.exists(path):
                continue
            for root, dirs, files in os.walk(path):
                root = normpath(root)
                out[root] = os.path.basename(root)
                
                for d in dirs:
                    if d.startswith(('.', '__')):
                        continue
                    if '-' in d or '.' in d:
                        continue
                    out[f'{root}/{d}'] = d
                
                # for f in files:
                #     name, ext = os.path.splitext(f)
                #     if ext not in ('.py', '.pyc', '.pyd', '.pyo', '.pyw'):
                #         continue
                #     if '-' in name or '.' in name:
                #         continue
                #     out[f'{root}/{f}'] = name
                break
        
        return out
    
    # -------------------------------------------------------------------------
    
    def relpath(self, abs_path: str) -> str:
        out = normpath(os.path.relpath(abs_path, self._cwd))
        if not out.startswith('../'):
            out = './' + out
        return out
    
    def reformat_external_lib_path(self, path: str, style: str) -> str:
        """
        args:
            style:
                'relpath'
                'pretty_relpath'
                'lib_name_only'
        """
        if style == 'relpath':
            return self.relpath(path)
        else:
            for lib_path in sorted(self.external_libs, reverse=True):
                if path.startswith(lib_path):
                    lib_name = self.external_libs[lib_path].lower()
                    lib_relpath = path[len(lib_path):].lstrip('./') or \
                                  os.path.basename(path)
                    break
            else:
                lib_name = 'unknown'
                lib_relpath = path.lstrip('./')
            
            if style == 'pretty_relpath':
                return '[{}]/{}'.format(lib_name, lib_relpath)
            elif style == 'lib_name_only':
                return '[{}]'.format(lib_name)


path_helper = PathHelper()



from . import console
from ._print import builtin_print
from ._print import builtin_print as bprint  # alias
from .control import disable
from .control import enable
from .control import setup
from .control import start_ipython
from .control import unload
from .control import update
from .logger import lk

__version__ = '5.3.2'



from __future__ import annotations

import typing as t

from .markup import MarkMeaning
from .message_formatter import MessageFormatter


class T:
    from .markup import T as _TMarkup  # noqa
    Args = t.Tuple[t.Any, ...]
    Markup = _TMarkup.Markup
    MarksMeaning = _TMarkup.MarksMeaning


class MessageBuilder:
    
    def __init__(self, **kwargs):
        self._formatter = MessageFormatter()
        self.update_config(**kwargs)
    
    # noinspection PyAttributeOutsideInit
    def update_config(self, **config):
        self._separator = config.get('separator', ';   ')
        self._show_source = config.get('show_source', False)
        self._show_funcname = config.get('show_funcname', False)
        self._show_varnames = config.get('show_varnames', False)
    
    # -------------------------------------------------------------------------
    
    def quick_compose(self, args: T.Args) -> str:
        return (
            '[bright_black]{separator}[/]'.format(
                separator=self._separator
            ).join(map(str, args))
        )
    
    def compose(self,
                args: T.Args,
                marks_meaning: T.MarksMeaning,
                info: dict) -> str:
        if MarkMeaning.AGRESSIVE_PRUNE in marks_meaning:
            return self.quick_compose(args)
        
        message_elements = []
        
        # 1. source
        if self._show_source:
            message_elements.append(
                self._formatter.fmt_source(
                    info['file_path'],
                    info['line_number'],
                    is_external_lib=info['is_external_lib'],
                    fmt_width=True,
                )
            )
            message_elements.append(
                self._formatter.fmt_separator('  >>  ')
            )
        
        # 2. funcname
        if self._show_funcname:
            assert info['function_name']
            message_elements.append(
                self._formatter.fmt_funcname(
                    info['function_name'],
                    fmt_width=True,
                )
            )
            message_elements.append(
                self._formatter.fmt_separator('  >>  ')
            )
        
        # 3. verbosity
        if MarkMeaning.VERBOSITY in marks_meaning:
            if MarkMeaning.MODERATE_PRUNE not in marks_meaning:
                message_elements.append(
                    self._formatter.fmt_level(
                        marks_meaning[MarkMeaning.VERBOSITY],
                        text='{TAG} '
                    )
                )
        
        # 4. index
        if MarkMeaning.RESET_INDEX in marks_meaning:
            if MarkMeaning.MODERATE_PRUNE in marks_meaning:
                pass
            else:
                message_elements.append(self._formatter.fmt_index(0))
                message_elements.append(' ')
                if not args:
                    args = ('[grey50]reset index[/]',)
        elif MarkMeaning.UPDATE_INDEX in marks_meaning:
            message_elements.append(
                self._formatter.fmt_index(
                    marks_meaning[MarkMeaning.UPDATE_INDEX]
                )
            )
            message_elements.append(' ')
        
        # 5. timestamp
        if MarkMeaning.RESET_TIMER in marks_meaning:
            if not args:
                args = ('[grey50]reset timer: [/] {}'.format(
                    self._formatter.fmt_time(
                        marks_meaning[MarkMeaning.RESET_TIMER],
                        color_s='green dim'
                    )
                ),)
            else:
                args = (self._formatter.fmt_time(
                    marks_meaning[MarkMeaning.RESET_TIMER],
                    color_s='green dim'
                ), *args)
        elif MarkMeaning.START_TIMER in marks_meaning:
            args = ('[cyan]start timer:[/] {}'.format(
                self._formatter.fmt_time(
                    marks_meaning[MarkMeaning.RESET_TIMER]
                )
            ), *args)
        elif MarkMeaning.STOP_TIMER in marks_meaning:
            s, e = marks_meaning[MarkMeaning.STOP_TIMER]
            args = (self._formatter.fmt_time(s, e), *args)
        
        # 6. divider
        if MarkMeaning.DIVIDER_LINE in marks_meaning:
            div = marks_meaning[MarkMeaning.DIVIDER_LINE]
            message_elements.append(
                self._formatter.fmt_divider(div)
            )
            message_elements.append(' ')
        
        # 7. arguments
        message_elements.append(
            self._formatter.fmt_message(
                arguments=args,
                varnames=info['variable_names'] if self._show_varnames else (),
                rich=MarkMeaning.RICH_FORMAT in marks_meaning,
                expand=MarkMeaning.EXPAND_MULTIPLE_LINES in marks_meaning,
                separator=self._separator,
            )
        )
        if MarkMeaning.VERBOSITY in marks_meaning and \
                MarkMeaning.EXPAND_MULTIPLE_LINES not in marks_meaning:
            message_elements[-1] = self._formatter.fmt_level(
                marks_meaning[MarkMeaning.VERBOSITY],
                text=message_elements[-1]
            )
        
        return ''.join(message_elements)
